"""Tests for aio module."""
from unittest.mock import Mock, call, patch, sentinel

import grpc
from google.protobuf.empty_pb2 import Empty
from grpc.aio import Channel
from testfixtures import LogCapture

from frgal.aio import AsyncGrpcClient
from frgal.constants import GRPC_MAX_RECONNECT_BACKOFF_MS, GRPC_MIN_RECONNECT_BACKOFF_MS
from frgal.decoder import GrpcDecoder
from frgal.exceptions import ServiceNotFound, UnknownError
from frgal.utils import AsyncMock

from .proto.example_pb2 import Reply
from .utils import make_awaitable

try:
    from unittest import IsolatedAsyncioTestCase as TestCase  # type: ignore
except ImportError:  # pragma: no cover
    from asynctest import TestCase  # type: ignore

try:
    from grpc._channel import _MultiThreadedRendezvous as _Rendezvous
except ImportError:  # pragma: no cover
    from grpc._channel import _Rendezvous


class AsyncGrpcClientTest(TestCase):
    """Test AsyncGrpcClient."""

    def setUp(self):
        self.grpc_module_mock = Mock(spec=('coffeeStub', ))
        self.channel_mock = AsyncMock(spec=Channel)
        self.channel_options = (
            ('grpc.min_reconnect_backoff_ms', GRPC_MIN_RECONNECT_BACKOFF_MS),
            ('grpc.max_reconnect_backoff_ms', GRPC_MAX_RECONNECT_BACKOFF_MS),
        )

        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock):
            self.client = AsyncGrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
            self.client.channel

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_connect_insecure(self):
        client = AsyncGrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])
        with patch('grpc.aio.insecure_channel', return_value=sentinel.channel) as mock:
            client.channel

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(mock.mock_calls, [call(sentinel.netloc, options=self.channel_options)])

    def test_connect_secure(self):
        client = AsyncGrpcClient(
            sentinel.netloc,
            sentinel.decoder,
            grpc_modules=[self.grpc_module_mock],
            credentials=sentinel.credentials,
        )
        with patch('grpc.aio.secure_channel', return_value=sentinel.channel) as mock:
            client.channel

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(mock.mock_calls, [call(sentinel.netloc, sentinel.credentials, options=self.channel_options)])

    def test_get_stub(self):
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(self.channel_mock)])

    def test_get_stub_cache(self):
        # Test stubs are cached when using the same channel
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        # Get stub once
        self.client.get_stub('coffee')
        # Get stub twice
        self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)
        # Check it was created only once
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(self.channel_mock)])

    def test_get_stub_unknown(self):
        with self.assertRaisesRegex(ServiceNotFound, 'Service teaStub could not be found'):
            self.client.get_stub('tea')

    async def test_call(self):
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_empty(self):
        reply = Empty()
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), None)

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned '),
        )

    async def test_call_error(self):
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            await self.client.call('coffee', 'make', Empty())

    async def test_call_grpc_unavailable(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = [error, error, error, make_awaitable(reply)]

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_grpc_unavailable_too_many_times(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            with self.assertRaises(grpc.RpcError):
                await self.client.call('coffee', 'make', Empty(), max_retries=3)

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Retries were unsuccessful.'),
            ('frgal.aio', 'INFO',
             '[XXXX] make failed with <_MultiThreadedRendezvous of RPC that terminated with:\n'
             '\tstatus = StatusCode.UNAVAILABLE\n'
             '\tdetails = "Connection reset by peer"\n'
             '\tdebug_error_string = "None"\n'
             '>')
        )

    async def test_call_grpc_other_error(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock) as channel_mock:
            with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
                await self.client.call('coffee', 'make', Empty())

        self.assertEqual(channel_mock.mock_calls, [])
